/*	Dynamic Random Dithering proof-of-concept program	*/
/* 		2006 Michel Rouzic aka A_SN			*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <png.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>

#define IMAGE_COUNT 7
#define FRAME_COUNT 200
#define RAND_COUNT 65535

__inline uint8_t upperpower(int32_t n)
{
	uint8_t p=0;

	while (n!=0)
	{
		n>>=1;
		p++;
	}

	return p;
}

//"stolen" from XBubble's source code and slightly modified to my liking
uint8_t *load_png_file(const char *file, uint16_t *width, uint16_t *height, uint8_t *i_power, uint8_t mode)
{
	/* Modes :
	 * 0 : RGB888 w/ power
	 * 1 : RGB888
	 * 2 : 8-bit grey
	 */

	FILE *fd;
	uint8_t *data, *image, header[8], ic;
	int32_t  bit_depth, color_type, i, iy, ix;
	png_uint_32  png_width, png_height, rowbytes;
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep *row_pointers;

	fd = fopen( file, "rb" );
	if ( fd == NULL )
	{
		perror(file);
		return NULL;
	}
	// ensure that we opened a PNG file
	fread( header, 1, 8, fd );
	if ( ! png_check_sig( header, 8 ) )
	{
		fclose(fd);
		printf("File %s does not have a valid PNG signature.\n", file);
		return NULL;
	}
	// create PNG structs
	png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if ( ! png_ptr )
	{
		fclose(fd);
		return NULL;
	}
	info_ptr = png_create_info_struct(png_ptr);
	if ( ! info_ptr )
	{
		png_destroy_read_struct( &png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		fclose(fd);
		return NULL;
	}
	//libpng does a longjmp here when it encounters an error
	if ( setjmp( png_ptr->jmpbuf ) )
	{
		png_destroy_read_struct( &png_ptr, &info_ptr, NULL);
		fclose(fd);
		return NULL;
	}
	png_init_io( png_ptr, fd );
	png_set_sig_bytes( png_ptr, 8);
	png_read_info( png_ptr, info_ptr);
	png_get_IHDR( png_ptr, info_ptr, &png_width, &png_height, &bit_depth,
	              &color_type, NULL, NULL, NULL);
	*width = (int16_t) png_width;
	*height = (int16_t) png_height;

/*********************SHOULDN'T BE NEEDED****************************/
	//convert image format to RGB(A)
	if ( color_type == PNG_COLOR_TYPE_PALETTE )
		png_set_expand(png_ptr);

	if (( color_type == PNG_COLOR_TYPE_GRAY ) ||
	        ( color_type == PNG_COLOR_TYPE_GRAY_ALPHA ))
		png_set_gray_to_rgb(png_ptr);

	// we only handle 8 bits per channel
	if (bit_depth == 16)
		png_set_strip_16(png_ptr);
	// insert filling bytes if necessary
	png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);

	// update info data
	png_read_update_info( png_ptr, info_ptr);
/********************************************************************/

	// allocate space for data and row pointers
	rowbytes = png_get_rowbytes( png_ptr, info_ptr);
	data = malloc (rowbytes * (*height));
	row_pointers = malloc ((*height) * sizeof(png_bytep));
	if (( data == NULL ) || ( row_pointers == NULL ))
	{
		png_destroy_read_struct( &png_ptr, &info_ptr, NULL);
		free(data);
		free(row_pointers);
		return NULL;
	}
	// initialize row pointers
	for ( i = 0;  i < *height; i++ )
		row_pointers[i] = data + i * rowbytes;
	// read all image data
	png_read_image( png_ptr, row_pointers );
	png_read_end( png_ptr, NULL);
	// cleanup
	free(row_pointers);
	png_destroy_read_struct( &png_ptr, &info_ptr, NULL);
	fclose(fd);
	// return an array of RGB(A) values
	//return data;

	*i_power=upperpower(*width);

	if (mode==0)	//RGB w/ power
	{
		image = malloc ((1<<*i_power) * (*height) * 4 * sizeof(uint8_t));
		for (iy=0; iy<*height; iy++)
			for (ix=0; ix<(1<<*i_power); ix++)
				for (ic=0; ic<3; ic++)
					image[(((iy*((1<<*i_power)))+ix)<<2)+ic] = data[(iy * (*width) + ix) * 4 + ic];
	}

	if (mode==1)	//RGB
	{
		image = malloc ((*width) * (*height) * 4 * sizeof(uint8_t));
		for (iy=0; iy<*height; iy++)
			for (ix=0; ix<*width; ix++)
				for (ic=0; ic<3; ic++)
					image[(((iy*(*width))+ix)<<2)+ic] = data[(iy * (*width) + ix) * 4 + ic];
	}

	if (mode==2)	//8-bit grey
	{
		image = malloc ((*width) * (*height) * sizeof(uint8_t));
		for (iy=0; iy<*height; iy++)
			for (ix=0; ix<*width; ix++)
				image[iy * (*width) + ix] =  data[(iy * (*width) + ix) * 4];
	}

	free(data);
	return image;
}

void dither(uint8_t *noise, uint32_t noise_size, uint8_t *image, uint8_t *screen32, uint16_t width, uint16_t height)
{
	static uint32_t in=0;
	uint32_t i;
	uint16_t p;

	/*for (i=(width*height-1)*4; i!=0; i-=4)
	{
		p = image[i] + rand()%16;	//RED (5 bits)
		if (p>255)
			p=255;
		screen32[i] = p;
		p = image[i+1] + rand()%8;	//GREEN (6 bits)
		if (p>255)
			p=255;
		screen32[i+1] = p;
		p = image[i+2] + rand()%16;	//BLUE (5 bits)
		if (p>255)
			p=255;
		screen32[i+2] = p;
	}*/

	for (i=(width*height-1)*4; i!=0; i-=4)
	{
		if (in==noise_size-3)
			in=0;
		else
			in+=3;
		p = image[i] + noise[in];	//RED
		if (p>255)
			p=255;
		screen32[i] = p;

		p = image[i+1] + noise[in+1];	//GREEN
		if (p>255)
			p=255;
		screen32[i+1] = p;

		p = image[i+2] + noise[in+2];	//BLUE
		if (p>255)
			p=255;
		screen32[i+2] = p;
	}
}

extern void fb_write(uint64_t *screen, uint32_t *fb);	//FIXME

int main(int argc, char *argv[])
{
	int fbdev;
	uint32_t i, j;
	uint16_t *fb, image_w, image_h;
	uint8_t *screen32, *image, *noise, trash1;
	char *image_name[IMAGE_COUNT];
	
	fbdev = open("/dev/fb0", O_RDWR);
	fb = (uint16_t *) mmap(0, 320*240*sizeof(uint16_t), PROT_WRITE, MAP_SHARED, fbdev, 0);
	screen32 = calloc (320 * 240 * 4, sizeof(uint8_t));

	//image_name[0] = "0.png";

	for (i=0; i<IMAGE_COUNT; i++)
	{
		image_name[i] = calloc (8, sizeof(char));
		sprintf(image_name[i], "%d.png", i);
	}

	noise = malloc (RAND_COUNT * sizeof(uint8_t));
	for (i=0; i<RAND_COUNT; i+=3)
	{
		noise[i] = rand()%16;
		noise[i+1] = rand()%8;
		noise[i+2] = rand()%16;
	}

	for (i=0; i<IMAGE_COUNT; i++)
	{
		image = load_png_file(image_name[i], &image_w, &image_h, &trash1, 1);
		if (image_w!=320 || image_h!=240)
			printf("WARNING : Wrong dimensions for image %s\n", image_name[i]);
		for (j=0; j<FRAME_COUNT; j++)
			fb_write((uint64_t *)image, (uint32_t *)fb);	//undithered display
		for (j=0; j<FRAME_COUNT; j++)
		{
			dither(noise, RAND_COUNT, image, screen32, 320, 240);
			fb_write((uint64_t *)screen32, (uint32_t *)fb);	//dithered display
		}
		free(image);
	}

	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);

	return 0;
}
